from redpic.utils.field import (
    get_field_accelerator,
    get_field_beam,
    sum_field_particles,
)
from redpic.utils.jit import jit

__all__ = ["get_field_beam", "get_field_accelerator", "sum_field_particles", "jit"]



import logging
from abc import ABC, abstractmethod
from collections.abc import Iterable

from redpic import constants as const
from redpic.accelerator import Accelerator
from redpic.beam.base import BaseBeam

module_logger = logging.getLogger(__name__)


class BaseSimulation(ABC):
    def __init__(
        self,
        beam: BaseBeam | Iterable[BaseBeam],
        accelerator: Accelerator,
        *,
        t_start: float = 0.0e0,
        t_stop: float = 0.0e0,
        dt: float = 0.0e0,
    ):
        self.beams = beam if isinstance(beam, Iterable) else [beam]
        self.acc = accelerator
        self.result = {}
        self.t_start = t_start
        self.t_stop = t_stop if t_stop else (self.acc.z_stop - self.acc.z_start) / const.c
        self.dt = dt if dt else self.acc.dz / const.c

    @abstractmethod
    def _track(self, *, n_files: int) -> None:
        pass

    def track(self, *, n_files: int = 20) -> None:
        if n_files <= 0:
            raise ValueError("The number of files (n_files) must be a positive number!")

        try:
            self._track(n_files=n_files)
        except KeyboardInterrupt:
            module_logger.info("Simulation interrupted by user")



LOG_FORMAT = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

LOG_DEFAULT_HANDLERS = [
    "console",
]

LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "formatters": {
        "verbose": {"format": LOG_FORMAT},
    },
    "handlers": {
        "console": {
            "level": "DEBUG",
            "class": "logging.StreamHandler",
            "formatter": "verbose",
        },
    },
    "loggers": {
        "": {
            "handlers": LOG_DEFAULT_HANDLERS,
            "level": "INFO",
        },
    },
    "root": {
        "level": "INFO",
        "formatter": "verbose",
        "handlers": LOG_DEFAULT_HANDLERS,
    },
}



import numpy as np
import pandas as pd
from numba import cuda, prange

from redpic import constants as const
from redpic.core import config as cfg
from redpic.solver.base import BaseSimulation
from redpic.utils.field import get_field_accelerator, get_field_beam
from redpic.utils.jit import jit


class REDSimulation(BaseSimulation):
    @staticmethod
    @jit
    def _first_step_red(
        x: np.array,
        y: np.array,
        z: np.array,
        px: np.array,
        py: np.array,
        pz: np.array,
        Fx: np.array,
        Fy: np.array,
        Fz: np.array,
        z_start: float,
        z_stop: float,
    ) -> None:
        for i in prange(int(len(x))):  # pylint: disable=E1133
            if z_start <= z[i] <= z_stop:
                px[i] += 2 * Fx[i]
                py[i] += 2 * Fy[i]
                pz[i] += 2 * Fz[i]
                gamma = (1 + px[i] ** 2 + py[i] ** 2 + pz[i] ** 2) ** (1 / 2)
                vx = px[i] / gamma
                vy = py[i] / gamma
                vz = pz[i] / gamma
                x[i] += vx
                y[i] += vy
                z[i] += vz
            else:
                gamma = (1 + px[i] ** 2 + py[i] ** 2 + pz[i] ** 2) ** (1 / 2)
                vz = pz[i] / gamma
                z[i] += vz

    @staticmethod
    @cuda.jit
    def _first_step_red_cuda(
        x: np.array,
        y: np.array,
        z: np.array,
        px: np.array,
        py: np.array,
        pz: np.array,
        Fx: np.array,
        Fy: np.array,
        Fz: np.array,
        z_start: float,
        z_stop: float,
    ) -> None:
        i = cuda.grid(1)  # pylint: disable=no-value-for-parameter
        if z_start <= z[i] <= z_stop and i < x.size:
            px[i] += 2 * Fx[i]
            py[i] += 2 * Fy[i]
            pz[i] += 2 * Fz[i]
            gamma = (1 + px[i] ** 2 + py[i] ** 2 + pz[i] ** 2) ** (1 / 2)
            vx = px[i] / gamma
            vy = py[i] / gamma
            vz = pz[i] / gamma
            x[i] += vx
            y[i] += vy
            z[i] += vz
        else:
            gamma = (1 + px[i] ** 2 + py[i] ** 2 + pz[i] ** 2) ** (1 / 2)
            vz = pz[i] / gamma
            z[i] += vz

    @staticmethod
    @jit
    def _second_step_red(
        x: np.array,
        y: np.array,
        z: np.array,
        px: np.array,
        py: np.array,
        pz: np.array,
        Fx: np.array,
        Fy: np.array,
        Fz: np.array,
        z_start: float,
        z_stop: float,
    ) -> None:
        for i in prange(int(len(x))):  # pylint: disable=E1133
            if z_start <= z[i] <= z_stop:
                gamma = (1 + px[i] ** 2 + py[i] ** 2 + pz[i] ** 2) ** (1 / 2)
                vx = px[i] / gamma
                vy = py[i] / gamma
                vz = pz[i] / gamma
                b2 = 1 + Fx[i] ** 2 + Fy[i] ** 2 + Fz[i] ** 2
                b1 = 2 - b2
                b3 = 2 * (vx * Fx[i] + vy * Fy[i] + vz * Fz[i])
                fx = 2 * (vy * Fz[i] - vz * Fy[i])
                fy = 2 * (vz * Fx[i] - vx * Fz[i])
                fz = 2 * (vx * Fy[i] - vy * Fx[i])
                vx = (vx * b1 + fx + Fx[i] * b3) / b2
                vy = (vy * b1 + fy + Fy[i] * b3) / b2
                vz = (vz * b1 + fz + Fz[i] * b3) / b2
                x[i] += vx
                y[i] += vy
                z[i] += vz
                px[i] = vx * gamma
                py[i] = vy * gamma
                pz[i] = vz * gamma
            else:
                gamma = (1 + px[i] ** 2 + py[i] ** 2 + pz[i] ** 2) ** (1 / 2)
                vz = pz[i] / gamma
                z[i] += vz

    @staticmethod
    @cuda.jit
    def _second_step_red_cuda(
        x: np.array,
        y: np.array,
        z: np.array,
        px: np.array,
        py: np.array,
        pz: np.array,
        Fx: np.array,
        Fy: np.array,
        Fz: np.array,
        z_start: float,
        z_stop: float,
    ) -> None:
        i = cuda.grid(1)  # pylint: disable=no-value-for-parameter
        if z_start <= z[i] <= z_stop and i < x.size:
            gamma = (1 + px[i] ** 2 + py[i] ** 2 + pz[i] ** 2) ** (1 / 2)
            vx = px[i] / gamma
            vy = py[i] / gamma
            vz = pz[i] / gamma
            b2 = 1 + Fx[i] ** 2 + Fy[i] ** 2 + Fz[i] ** 2
            b1 = 2 - b2
            b3 = 2 * (vx * Fx[i] + vy * Fy[i] + vz * Fz[i])
            fx = 2 * (vy * Fz[i] - vz * Fy[i])
            fy = 2 * (vz * Fx[i] - vx * Fz[i])
            fz = 2 * (vx * Fy[i] - vy * Fx[i])
            vx = (vx * b1 + fx + Fx[i] * b3) / b2
            vy = (vy * b1 + fy + Fy[i] * b3) / b2
            vz = (vz * b1 + fz + Fz[i] * b3) / b2
            x[i] += vx
            y[i] += vy
            z[i] += vz
            px[i] = vx * gamma
            py[i] = vy * gamma
            pz[i] = vz * gamma
        else:
            gamma = (1 + px[i] ** 2 + py[i] ** 2 + pz[i] ** 2) ** (1 / 2)
            vz = pz[i] / gamma
            z[i] += vz

    def _track(self, *, n_files: int = cfg.DEFAULT_TRACK_SAVE_N_FILES) -> None:
        # Init parameters
        z_start, z_stop, dz = self.acc.z_start, self.acc.z_stop, self.acc.dz
        t_start, t_stop, dt = self.t_start, self.t_stop, self.dt

        Y = np.transpose(np.concatenate([beam.df.to_numpy() for beam in self.beams]))
        Y[2] = Y[2] + z_start - max(Y[2])  # set initial beam position

        m = np.transpose(np.concatenate([np.full(beam.n, beam.type.mass) for beam in self.beams]))
        phys_q = np.transpose(np.concatenate([np.full(beam.n, beam.type.charge) for beam in self.beams]))
        macro_q = np.transpose(np.concatenate([np.full(beam.n, beam.total_charge / beam.n) for beam in self.beams]))

        P0 = m * const.c * const.c / (const.e * const.mega)
        E0 = m * const.c / (phys_q * dt * const.mega)
        B0 = m / (phys_q * dt)

        # RED
        for t in np.arange(t_start, t_stop, 2 * dt):
            # into the internal system
            Y[0], Y[1], Y[2] = Y[0] / dz, Y[1] / dz, Y[2] / dz
            Y[3], Y[4], Y[5] = Y[3] / P0, Y[4] / P0, Y[5] / P0

            # get electric field from accelerator
            Ex, Ey, Ez = get_field_accelerator(self.acc, "E", Y[0] * dz, Y[1] * dz, Y[2] * dz)
            Ex, Ey, Ez = Ex / E0, Ey / E0, Ez / E0

            # get electric field from beam
            if bool(beam.total_charge for beam in self.beams):
                ex, ey, ez = get_field_beam(macro_q, self.acc, "E", Y[0] * dz, Y[1] * dz, Y[2] * dz)
                ex, ey, ez = ex / E0, ey / E0, ez / E0
                Ex, Ey, Ez = Ex + ex, Ey + ey, Ez + ez

            # first step RED
            if cfg.ENABLE_CUDA:
                self._first_step_red_cuda[
                    min(beam.n for beam in self.beams) // cfg.CUDA_THREADS_PER_BLOCK + 1, cfg.CUDA_THREADS_PER_BLOCK
                ](Y[0], Y[1], Y[2], Y[3], Y[4], Y[5], Ex, Ey, Ez, z_start / dz, z_stop / dz)
            else:
                self._first_step_red(Y[0], Y[1], Y[2], Y[3], Y[4], Y[5], Ex, Ey, Ez, z_start / dz, z_stop / dz)
            gamma = np.sqrt(1 + Y[3] * Y[3] + Y[4] * Y[4] + Y[5] * Y[5])
            vz = Y[5] / gamma

            # get magnetic field from accelerator
            Bx, By, Bz = get_field_accelerator(self.acc, "B", Y[0] * dz, Y[1] * dz, Y[2] * dz)
            Bx, By, Bz = Bx / B0 / gamma, By / B0 / gamma, Bz / B0 / gamma

            # get magnetic field from beam
            if bool(beam.total_charge for beam in self.beams):
                bx, by, bz = get_field_beam(macro_q, self.acc, "B", Y[0] * dz, Y[1] * dz, Y[2] * dz)
                bx, by, bz = bx * vz / B0, by * vz / B0, bz * vz / B0
                Bx, By, Bz = Bx + bx, By + by, Bz + bz

            # second step RED
            if cfg.ENABLE_CUDA:
                self._second_step_red_cuda[
                    min(beam.n for beam in self.beams) // cfg.CUDA_THREADS_PER_BLOCK + 1, cfg.CUDA_THREADS_PER_BLOCK
                ](Y[0], Y[1], Y[2], Y[3], Y[4], Y[5], Bx, By, Bz, z_start / dz, z_stop / dz)
            else:
                self._second_step_red(Y[0], Y[1], Y[2], Y[3], Y[4], Y[5], Bx, By, Bz, z_start / dz, z_stop / dz)
            gamma = np.sqrt(1 + Y[3] * Y[3] + Y[4] * Y[4] + Y[5] * Y[5])
            vz = Y[5] / gamma

            # into the SI
            Y[0], Y[1], Y[2] = Y[0] * dz, Y[1] * dz, Y[2] * dz
            Y[3], Y[4], Y[5] = Y[3] * P0, Y[4] * P0, Y[5] * P0
            Bx, By, Bz = Bx * gamma * B0, By * gamma * B0, Bz * gamma * B0
            Ex, Ey, Ez = Ex * E0, Ey * E0, Ez * E0

            progress, meters = t / t_stop * 100, z_start + t * const.c
            if progress % (100 // n_files) < 2 * dt / t_stop * 100:
                self.result.update(
                    {
                        round(meters, 3): pd.DataFrame(
                            np.transpose(
                                np.vstack((Y[0], Y[1], Y[2], Y[3], Y[4], Y[5], Bx, By, Bz, Ex, Ey, Ez, macro_q))
                            ),
                            columns=["x", "y", "z", "px", "py", "pz", "Bx", "By", "Bz", "Ex", "Ey", "Ez", "charge"],
                        )
                    }
                )
            print("\rz = %.2f m (%.1f %%) " % (meters, progress), end="")



import numpy as np
from scipy import integrate, interpolate


class Element:
    """
    Sets one accelerator element.

    """

    def __init__(
        self,
        z0: float,
        max_field: float,
        file_name: str,
        name: str,
        *,
        x: float = 0.0,
        xp: float = 0.0,
        y: float = 0.0,
        yp: float = 0.0,
    ):
        """
        Initialization of an accelerator element.

        Parameters
        ----------
        z0: float
            Position of the element center
        max_field: float
            Maximum value of the field on the element axis
        file_name: string
            Name of the file
        name: string
            Name of the element

        x: float, optional
            Offset of the element along the x-axis, [m]
        xp: float, optional
            Element rotation in the z-x plane, [rad]
        y: float, optional
            Offset of the element along the y-axis, [m]
        yp: float, optional
            Element rotation in the z-y plane, [rad]
        """
        self.z0 = z0
        self.max_field = max_field
        self.file_name = file_name
        self.name = name
        self.x = x
        self.xp = xp
        self.y = y
        self.yp = yp

        self.z_start = 0.0e0
        self.z_stop = 0.0e0
        self.length = 0.0e0
        self.field = 0.0e0


def read_fields(beamline: dict, z: np.arange) -> interpolate.interp1d:
    """
    Field reading function.

    Parameters
    ----------
    beamline: dict
        Accelerator`s beamline
    z: an instance of the `np.arange` class
        Axis z

    Returns
    -------
    F: an instance of the `interpolate.interp1d` class
        Field function
    F_prime: an instance of the `interpolate.interp1d` class
        Field prime function
    F_int: an instance of the `interpolate.interp1d` class
        Field integrate function
    """

    field_files = {}
    F, F_prime, F_int = 0, 0, 0

    if not beamline:
        z_data = [i / len(z) for i in range(len(z))]
        F_data = [0 for i in range(len(z))]
        f = interpolate.interp1d(z_data, F_data, fill_value=(0, 0), bounds_error=False)
        F = F + f(z)
        F_prime = F
        F_int = F
    else:
        for element in beamline.values():
            if not (element.file_name in field_files):  # pylint: disable=superfluous-parens
                field_files[element.file_name] = np.loadtxt(element.file_name)
            M = field_files[element.file_name]
            z_data = M[:, 0] + element.z0
            F_data = M[:, 1]
            dz = (z_data[-1] - z_data[0]) / len(z_data)

            f = interpolate.interp1d(
                z_data, element.max_field * F_data, kind="cubic", fill_value=(0, 0), bounds_error=False
            )

            f_prime = interpolate.interp1d(
                z_data, element.max_field * np.gradient(F_data, dz), kind="cubic", fill_value=(0, 0), bounds_error=False
            )

            F = F + f(z)
            F_prime = F_prime + f_prime(z)

            if not element.max_field == 0:
                int_f = integrate.cumulative_trapezoid(f(z_data), z_data)[-1]
                int_ff = integrate.cumulative_trapezoid(f(z_data) ** 2, z_data)[-1]
                element.length = int_f**2 / int_ff
                element.field = int_ff / int_f
                element.z_start = element.z0 - element.length
                element.z_stop = element.z0 + element.length
            else:
                element.length = 0
                element.field = 0
                element.z_start = element.z0
                element.z_stop = element.z0

    F = interpolate.interp1d(z, F, kind="cubic", fill_value=(0, 0), bounds_error=False)
    F_int = integrate.cumulative_trapezoid(F(z), z)
    F_prime = interpolate.interp1d(z, F_prime, kind="cubic", fill_value=(0, 0), bounds_error=False)
    F_int = interpolate.interp1d(z[1:], F_int, kind="cubic", fill_value=(F_int[0], F_int[-1]), bounds_error=False)

    return F, F_prime, F_int


def read_offsets(beamline: dict, z: np.arange) -> interpolate.interp1d:
    """
    Offsets and rotations reading function.

    Parameters
    ----------
    beamline: dict
        Accelerator`s beamline
    z: an instance of the `np.arange` class
        Axis z

    Returns
    -------
    offset_correct_x: an instance of the `interpolate.interp1d` class
        Offset function
    offset_correct_xp: an instance of the `interpolate.interp1d` class
        Rotation function
    offset_correct_y: an instance of the `interpolate.interp1d` class
        Offset function
    offset_correct_yp: an instance of the `interpolate.interp1d` class
        Rotation function
    """

    field_files = {}
    F = 0
    offset_correct_x, offset_correct_y = 0, 0
    offset_correct_xp, offset_correct_yp = 0, 0

    if not beamline:
        z_data = [i / len(z) for i in range(len(z))]
        F_data = [0 for i in range(len(z))]
        f = interpolate.interp1d(z_data, F_data, fill_value=(0, 0), bounds_error=False)
        F = F + f(z)
        offset_correct_x, offset_correct_y = F, F
        offset_correct_xp, offset_correct_yp = F, F
    else:
        for element in beamline.values():
            if not (element.file_name in field_files):  # pylint: disable=superfluous-parens
                field_files[element.file_name] = np.loadtxt(element.file_name)
            M = field_files[element.file_name]
            z_data = M[:, 0] + element.z0
            F_data = M[:, 1]

            if not element.length == 0:
                z_data = np.linspace(element.z_start, element.z_stop, len(z_data))
                x, xp = element.x, element.xp
                y, yp = element.y, element.yp
                z0 = element.z0
                f_x = interpolate.interp1d(
                    z_data,
                    [x + (z_data[i] - z0) * xp for i in range(len(z_data))],
                    fill_value=(0, 0),
                    bounds_error=False,
                )
                f_xp = interpolate.interp1d(
                    z_data, [xp for i in range(len(z_data))], fill_value=(0, 0), bounds_error=False
                )
                f_y = interpolate.interp1d(
                    z_data,
                    [y + (z_data[i] - z0) * yp for i in range(len(z_data))],
                    fill_value=(0, 0),
                    bounds_error=False,
                )
                f_yp = interpolate.interp1d(
                    z_data, [yp for i in range(len(z_data))], fill_value=(0, 0), bounds_error=False
                )
            else:
                f = interpolate.interp1d(z_data, 0 * F_data, kind="cubic", fill_value=(0, 0), bounds_error=False)
                f_x, f_xp, f_y, f_yp = f, f, f, f

            offset_correct_x = offset_correct_x + f_x(z)
            offset_correct_xp = offset_correct_xp + f_xp(z)
            offset_correct_y = offset_correct_y + f_y(z)
            offset_correct_yp = offset_correct_yp + f_yp(z)

    offset_correct_x = interpolate.interp1d(z, offset_correct_x, kind="linear", fill_value=(0, 0), bounds_error=False)
    offset_correct_y = interpolate.interp1d(z, offset_correct_y, kind="linear", fill_value=(0, 0), bounds_error=False)
    offset_correct_xp = interpolate.interp1d(z, offset_correct_xp, kind="linear", fill_value=(0, 0), bounds_error=False)
    offset_correct_yp = interpolate.interp1d(z, offset_correct_yp, kind="linear", fill_value=(0, 0), bounds_error=False)
    return (offset_correct_x, offset_correct_xp, offset_correct_y, offset_correct_yp)


class Accelerator:
    """
    Top-level accelerator class that contains all the accelerator data.

    The `Accelerator` class has several important attributes:

    - `Bx_beamline`, a beamline which contains the solenoids information
    - `By_beamline`, a beamline which contains the solenoids information
    - `Bz_beamline`, a beamline which contains the solenoids information
    - `Ez_beamline`, a beamline which contains the accels information
    - `Gz_beamline`, a beamline which contains the quadrupoles information
    """

    def __init__(self, z_start: float, z_stop: float, dz: float):
        """
        Initializes an accelerator.

        Parameters
        ----------
        z_start: float
            The position of the start of the accelerator
        z_stop: float
            The position of the end of the accelerator
        dz: float
            The step of the accelerator
        """
        self.z_start = self.start = z_start
        self.z_stop = self.stop = z_stop
        self.dz = self.step = dz
        self.z = self.parameter = np.arange(z_start, z_stop, dz)
        self.Bx_beamline, self.By_beamline, self.Bz_beamline = {}, {}, {}
        self.Ez_beamline = {}
        self.Gz_beamline = {}
        self.Bx, self.By = interpolate.interp1d, interpolate.interp1d
        self.Bz = interpolate.interp1d
        self.Ez = interpolate.interp1d
        self.Gz = interpolate.interp1d
        self.dBxdz, self.dBydz = interpolate.interp1d, interpolate.interp1d
        self.dBzdz = interpolate.interp1d
        self.dEzdz = interpolate.interp1d
        self.dGzdz = interpolate.interp1d
        self.Bxdz, self.Bydz = interpolate.interp1d, interpolate.interp1d
        self.Bzdz = interpolate.interp1d
        self.Ezdz = interpolate.interp1d
        self.Gzdz = interpolate.interp1d
        self.Dx, self.Dxp = interpolate.interp1d, interpolate.interp1d
        self.Dy, self.Dyp = interpolate.interp1d, interpolate.interp1d

    def add_solenoid(
        self,
        name: str,
        center: float,
        max_field: float,
        file_name: str,
        *,
        x: float = 0.0,
        xp: float = 0.0,
        y: float = 0.0,
        yp: float = 0.0,
    ) -> None:
        """
        Creates a solenoid lenses in the accelerator.

        Parameters
        ----------
        name: string
            Solenoid's id
        center: float
            Solenoid's center
        max_field: float
            Solenoid's maximum field [T]
        file_name: string
            Experimental profile of the Bz field

        x: float, optional
            Offset of the element along the x-axis
        xp: float, optional
            Element rotation in the z-x plane
        y: float, optional
            Offset of the element along the y-axis
        yp: float, optional
            Element rotation in the z-y plane
        """
        self.Bz_beamline[name] = Element(center, max_field, file_name, name, x=x, xp=xp, y=y, yp=yp)

    def add_accel(
        self,
        name: str,
        center: float,
        max_field: float,
        file_name: str,
        *,
        x: float = 0.0,
        xp: float = 0.0,
        y: float = 0.0,
        yp: float = 0.0,
    ) -> None:
        """
        Creates an accelerating module in the accelerator.

        Parameters
        ----------
        name: string
            Accel's id
        center: float
            Accel's center
        max_field: float
            Accel's maximum field [MV/m]
        file_name: string
            Experimental profile of the Ez field

        x: float, optional
            Offset of the element along the x-axis
        xp: float, optional
            Element rotation in the z-x plane
        y: float, optional
            Offset of the element along the y-axis
        yp: float, optional
            Element rotation in the z-y plane
        """
        self.Ez_beamline[name] = Element(center, max_field, file_name, name, x=x, xp=xp, y=y, yp=yp)

    def add_quadrupole(self, name: str, center: float, max_field: float, file_name: str) -> None:
        """
        Creates a quadrupole in the accelerator.

        Parameters
        ----------
        name: string
            Quad's id
        center: float
            Quad's center
        max_field: float
            Quad's maximum field [T/m]
        file_name: string
            Experimental profile of the Gz field

        x: float, optional
            Offset of the element along the x-axis
        xp: float, optional
            Element rotation in the z-x plane
        y: float, optional
            Offset of the element along the y-axis
        yp: float, optional
            Element rotation in the z-y plane
        """
        self.Gz_beamline[name] = Element(center, max_field, file_name, name)

    def add_corrector_x(self, name: str, center: float, max_field: float, file_name: str) -> None:
        """
        Creates a corrector x in the accelerator.

        Parameters
        ----------
        name: string
            Corrector's id
        center: float
            Corrector's center
        max_field: float
            Corrector's maximum field [T]
        file_name: string
            Experimental profile of the By field

        x: float, optional
            Offset of the element along the x-axis
        xp: float, optional
            Element rotation in the z-x plane
        y: float, optional
            Offset of the element along the y-axis
        yp: float, optional
            Element rotation in the z-y plane
        """
        self.By_beamline[name] = Element(center, max_field, file_name, name)

    def add_corrector_y(self, name: str, center: float, max_field: float, file_name: str) -> None:
        """
        Creates a corrector y in the accelerator.

        Parameters
        ----------
        name: string
            Corrector's id
        center: float
            Corrector's center
        max_field: float
            Corrector's maximum field [T]
        file_name: string
            Experimental profile of the Bx field

        x: float, optional
            Offset of the element along the x-axis
        xp: float, optional
            Element rotation in the z-x plane
        y: float, optional
            Offset of the element along the y-axis
        yp: float, optional
            Element rotation in the z-y plane
        """
        self.Bx_beamline[name] = Element(center, max_field, file_name, name)

    def delete_solenoid(self, name: str = "all") -> None:
        """
        Delete a solenoid in the accelerator.

        Parameters
        ----------
        name: string
            Solenoid`s id
            (If there is no name that will be removed all)
        """
        if name == "all":
            self.Bz_beamline = {}
        else:
            self.Bz_beamline.pop(name)

    def delete_accel(self, name: str = "all") -> None:
        """
        Delete an accelerating module in the accelerator.

        Parameters
        ----------
        name: string
            Accel`s id
            (If there is no name that will be removed all)
        """
        if name == "all":
            self.Ez_beamline = {}
        else:
            self.Ez_beamline.pop(name)

    def delete_quadrupole(self, name: str = "all") -> None:
        """
        Delete a quadrupole in the accelerator.

        Parameters
        ----------
        name: string
            Quad`s id
            (If there is no name that will be removed all)
        """
        if name == "all":
            self.Gz_beamline = {}
        else:
            self.Gz_beamline.pop(name)

    def delete_corrector_x(self, name: str = "all") -> None:
        """
        Delete a corrector x in the accelerator.

        Parameters
        ----------
        name: string
            Corrector`s id
            (If there is no name that will be removed all)
        """
        if name == "all":
            self.By_beamline = {}
        else:
            self.By_beamline.pop(name)

    def delete_corrector_y(self, name: str = "all") -> None:
        """
        Delete a corrector y in the accelerator.

        Parameters
        ----------
        name: string
           Corrector`s id
           (If there is no name that will be removed all)
        """
        if name == "all":
            self.Bx_beamline = {}
        else:
            self.Bx_beamline.pop(name)

    def compile(self) -> None:
        """
        Compilation of the accelerator.
        """

        self.Bx, self.dBxdz, self.Bxdz = read_fields(self.Bx_beamline, self.parameter)
        self.By, self.dBydz, self.Bydz = read_fields(self.By_beamline, self.parameter)
        self.Bz, self.dBzdz, self.Bzdz = read_fields(self.Bz_beamline, self.parameter)
        self.Ez, self.dEzdz, self.Ezdz = read_fields(self.Ez_beamline, self.parameter)
        self.Gz, self.dGzdz, self.Gzdz = read_fields(self.Gz_beamline, self.parameter)
        Dx_Bz, Dxp_Bz, Dy_Bz, Dyp_Bz = read_offsets(self.Bz_beamline, self.parameter)
        Dx_Ez, Dxp_Ez, Dy_Ez, Dyp_Ez = read_offsets(self.Ez_beamline, self.parameter)

        self.Dx = interpolate.interp1d(
            self.z, Dx_Bz(self.z) + Dx_Ez(self.z), kind="linear", fill_value=(0, 0), bounds_error=False
        )
        self.Dxp = interpolate.interp1d(
            self.z, Dxp_Bz(self.z) + Dxp_Ez(self.z), kind="linear", fill_value=(0, 0), bounds_error=False
        )
        self.Dy = interpolate.interp1d(
            self.z, Dy_Bz(self.z) + Dy_Ez(self.z), kind="linear", fill_value=(0, 0), bounds_error=False
        )
        self.Dyp = interpolate.interp1d(
            self.z, Dyp_Bz(self.z) + Dyp_Ez(self.z), kind="linear", fill_value=(0, 0), bounds_error=False
        )

    def __str__(self):
        string = "Accelerator structure.\n"
        string += "\tSolenoids:\n"
        for element in self.Bz_beamline.values():
            string += (
                f"\t[ {element.z0} m, {element.max_field} T, "
                f"{element.file_name}, {element.name}, "
                f"{element.x} m, {element.xp} rad, "
                f"{element.y} m, {element.yp} rad] \n"
            )
        string += "\tAccelerating modules:\n"
        for element in self.Ez_beamline.values():
            string += (
                f"\t[ {element.z0} m, {element.max_field} T, "
                f"{element.file_name}, {element.name}, "
                f"{element.x} m, {element.xp} rad, "
                f"{element.y} m, {element.yp} rad] \n"
            )
        string += "\tQuadrupoles:\n"
        for element in self.Gz_beamline.values():
            string += f"\t[ {element.z0} m, {element.max_field} T, " f"{element.file_name}, {element.name} \n"
        string += "\tCorrectors x:\n"
        for element in self.By_beamline.values():
            string += f"\t[ {element.z0} m, {element.max_field} T, " f"{element.file_name}, {element.name} \n"
        string += "\tCorrectors y:\n"
        for element in self.Bx_beamline.values():
            string += f"\t[ {element.z0} m, {element.max_field} T, " f"{element.file_name}, {element.name} \n"
        return string

    add_sol = add_new_solenoid = add_solenoid
    add_acc = add_new_accel = add_accel
    add_quad = add_new_quadrupole = add_quadrupole
    add_corr_x = add_new_corrector_x = add_corrector_x
    add_corr_y = add_new_corrector_y = add_corrector_y

    del_sol = del_solenoid = delete_solenoid
    del_acc = del_accel = delete_accel
    del_quad = del_quadrupole = delete_quadrupole
    del_corr_x = del_corrector_x = delete_corrector_x
    del_corr_y = del_corrector_y = delete_corrector_y



import logging

import numpy as np
import pandas as pd

from redpic.beam.base import BaseBeam, BeamDistributionType
from redpic.beam.uniform import get_4_sphere_normal

module_logger = logging.getLogger(__name__)


class RadialUniformBeam(BaseBeam):
    distribution = BeamDistributionType.RADIAL_UNIFORM

    def _generate(self, n: int = 0, *, file_name: str = "", **kwargs) -> None:
        dpz = kwargs.get("dpz", 0.01)
        self.n = n
        (s, t, u, v) = get_4_sphere_normal(self.n)
        (x, y) = (self.radius_x * s, self.radius_y * t)
        u = 2 * (self.px * self.normalized_emittance_x * u + self.radius_xp * x) / self.radius_x
        v = 2 * (self.py * self.normalized_emittance_y * v + self.radius_yp * y) / self.radius_y
        (x, y) = (x + self.x, y + self.y)
        (px, py) = (u + self.p * self.xp, v + self.p * self.yp)

        z = np.random.uniform(-self.radius_z + self.z, self.radius_z + self.z, int(self.n))
        pz = np.random.uniform((1 - dpz) * self.pz, (1 + dpz) * self.pz, int(self.n))

        self.df = pd.DataFrame(
            np.transpose(np.vstack((x, y, z, px, py, pz))), columns=["x", "y", "z", "px", "py", "pz"]
        )



import logging

import numpy as np
import pandas as pd

from redpic.beam.base import BaseBeam, BeamDistributionType

module_logger = logging.getLogger(__name__)


@staticmethod
def get_4_sphere_normal(n):
    s = np.random.normal(0, 1, int(n))
    t = np.random.normal(0, 1, int(n))
    u = np.random.normal(0, 1, int(n))  # pylint: disable=W0621
    v = np.random.normal(0, 1, int(n))
    norm = (s * s + t * t + u * u + v * v) ** 0.5
    return (s, t, u, v) / norm


class UniformBeam(BaseBeam):
    distribution = BeamDistributionType.UNIFORM

    def _generate(self, n: int = 0, *, file_name: str = "", **kwargs) -> None:
        module_logger.info("Generate a %s beam with %d particles", self.distribution, n)
        dpz = kwargs.get("dpz", 0.01)
        self.n = n
        (s, t, u, v) = get_4_sphere_normal(self.n)
        (x, y) = (self.radius_x * s + self.x, self.radius_y * t + self.y)
        (px, py) = (self.px * u + self.p * self.xp, v * self.py + self.p * self.yp)

        z = np.random.uniform(-self.radius_z + self.z, self.radius_z + self.z, int(self.n))
        pz = np.random.uniform((1 - dpz) * self.pz, (1 + dpz) * self.pz, int(self.n))

        self.df = pd.DataFrame(
            np.transpose(np.vstack((x, y, z, px, py, pz))), columns=["x", "y", "z", "px", "py", "pz"]
        )



from redpic.accelerator import *
from redpic.beam import *
from redpic.core import config
from redpic.solver import *

__version__ = config.VERSION

__doc__ = config.DOC

__all__ = accelerator.__all__ + beam.__all__ + solver.__all__



from numba import jit as numba_jit

from redpic.core import config as cfg


def jit(func):
    def wrapper(*args, **kwargs):
        if cfg.DISABLE_JIT:
            return func(*args, **kwargs)

        return numba_jit(
            nopython=not cfg.DISABLE_NOPYTHON,
            parallel=not cfg.DISABLE_PARALLEL,
            fastmath=not cfg.DISABLE_FASTMATH,
            cache=not cfg.DISABLE_CACHE,
            nogil=not cfg.DISABLE_NOGIL,
        )(func)(*args, **kwargs)

    return wrapper



from redpic.beam.astra import AstraBeam
from redpic.beam.base import BaseBeam
from redpic.beam.radial_uniform import RadialUniformBeam
from redpic.beam.uniform import UniformBeam

__all__ = ["BaseBeam", "UniformBeam", "RadialUniformBeam", "AstraBeam"]



import numpy as np
from numba import cuda, prange
from scipy import misc

from redpic import constants as const
from redpic.accelerator import Accelerator
from redpic.core import config as cfg
from redpic.utils.jit import jit


def get_field_accelerator(
    acc: Accelerator, type: str, x: np.array, y: np.array, z: np.array
) -> (np.array, np.array, np.array):
    dz = acc.dz
    offset_x = acc.Dx(z)
    offset_xp = acc.Dxp(z)
    offset_y = acc.Dy(z)
    offset_yp = acc.Dyp(z)
    x_corr = x - offset_x
    y_corr = y - offset_y
    r_corr = np.sqrt(x_corr * x_corr + y_corr * y_corr)
    if type == "E":
        Ez = acc.Ez(z)
        dEzdz = acc.dEzdz(z)
        d2Ezdz2 = misc.derivative(acc.dEzdz, z, dx=dz, n=1)
        Ez = Ez - d2Ezdz2 * r_corr * r_corr / 4  # row remainder
        Ex = -dEzdz * x_corr / 2 + Ez * offset_xp  # row remainder
        Ey = -dEzdz * y_corr / 2 + Ez * offset_yp  # row remainder
        return Ex, Ey, Ez
    if type == "B":
        Bx = acc.Bx(z)
        By = acc.By(z)
        Bz = acc.Bz(z)
        dBzdz = acc.dBzdz(z)
        d2Bzdz2 = misc.derivative(acc.dBzdz, z, dx=dz, n=1)
        Gz = acc.Gz(z)
        Bz = Bz - d2Bzdz2 * r_corr * r_corr / 4  # row remainder
        Bx = Bx + Gz * y_corr - dBzdz * x_corr / 2 + Bz * offset_xp  # row remainder
        By = By + Gz * x_corr - dBzdz * y_corr / 2 + Bz * offset_yp  # row remainder
        return Bx, By, Bz


@jit
def sum_field_particles(
    x: np.array, y: np.array, z: np.array, z_start: float, z_stop: float, Fx: np.array, Fy: np.array, Fz: np.array
) -> None:
    for i in prange(int(len(x))):  # pylint: disable=E1133
        if z_start <= z[i] <= z_stop:
            for j in range(int(len(x))):
                if z_start <= z[j] <= z_stop and i != j:
                    Fx[i] += (x[i] - x[j]) / ((x[j] - x[i]) ** 2 + (y[j] - y[i]) ** 2 + (z[j] - z[i]) ** 2) ** (3 / 2)
                    Fy[i] += (y[i] - y[j]) / ((x[j] - x[i]) ** 2 + (y[j] - y[i]) ** 2 + (z[j] - z[i]) ** 2) ** (3 / 2)
                    Fz[i] += (z[i] - z[j]) / ((x[j] - x[i]) ** 2 + (y[j] - y[i]) ** 2 + (z[j] - z[i]) ** 2) ** (3 / 2)


@cuda.jit
def sum_field_particles_cuda(
    x: np.array, y: np.array, z: np.array, z_start: float, z_stop: float, Fx: np.array, Fy: np.array, Fz: np.array
) -> None:
    i = cuda.grid(1)  # pylint: disable=no-value-for-parameter

    if i >= x.size:
        return

    for j in range(x.size):
        if z_start <= z[i] <= z_stop and z_start <= z[j] <= z_stop and i != j:
            Fx[i] += (x[i] - x[j]) / ((x[j] - x[i]) ** 2 + (y[j] - y[i]) ** 2 + (z[j] - z[i]) ** 2) ** (3 / 2)
            Fy[i] += (y[i] - y[j]) / ((x[j] - x[i]) ** 2 + (y[j] - y[i]) ** 2 + (z[j] - z[i]) ** 2) ** (3 / 2)
            Fz[i] += (z[i] - z[j]) / ((x[j] - x[i]) ** 2 + (y[j] - y[i]) ** 2 + (z[j] - z[i]) ** 2) ** (3 / 2)


def get_field_beam(
    q: np.array, acc: Accelerator, type: str, x: np.array, y: np.array, z: np.array
) -> (np.array, np.array, np.array):
    n = len(q)
    Fx, Fy, Fz = np.zeros(n), np.zeros(n), np.zeros(n)

    if cfg.ENABLE_CUDA:
        sum_field_particles_cuda[n // cfg.CUDA_THREADS_PER_BLOCK + 1, cfg.CUDA_THREADS_PER_BLOCK](
            x, y, z, acc.z_start, acc.z_stop, Fx, Fy, Fz
        )
    else:
        sum_field_particles(x, y, z, acc.z_start, acc.z_stop, Fx, Fy, Fz)

    if type == "E":
        k = const.ke / (4 * np.pi) * q
        return k * Fx * 1e-6, k * Fy * 1e-6, k * Fz * 1e-6
    if type == "B":
        k = const.km / (4 * np.pi) * q
        return k * Fx, -k * Fy, 0 * k * Fz



from logging import config as logging_config

from redpic.core.logger import LOGGING

logging_config.dictConfig(LOGGING)

NAME = "redpic"

VERSION = "0.11.7"

AUTHOR = "Vyacheslav Fedorov"

AUTHOR_EMAIL = "slava@fuodorov.ru"

DOC = "Relativistic Difference Scheme Particle-in-Cell code (REDPIC)"

URL = "https://github.com/fuodorov/redpic"

DISABLE_JIT = False

DISABLE_NOPYTHON = False

DISABLE_PARALLEL = False

DISABLE_FASTMATH = False

DISABLE_CACHE = False

DISABLE_NOGIL = False

ENABLE_CUDA = False

CUDA_THREADS_PER_BLOCK = 1024

DEFAULT_TRACK_SAVE_N_FILES = 20



import logging
from abc import ABC, abstractmethod

import numpy as np
import pandas as pd

from redpic import constants as const

module_logger = logging.getLogger(__name__)


class BeamDistributionType:
    NONE = "none"
    UNIFORM = "uniform"
    RADIAL_UNIFORM = "radial-uniform"
    GAUSSIAN = "gaussian"
    ASTRA = "astra"


class BaseBeam(ABC):
    distribution: BeamDistributionType = BeamDistributionType.NONE

    def __init__(
        self,
        *,
        type: const.Element = const.electron,
        current: float = 0.0e0,
        energy: float = 0.0e0,
        radius: float = 0.0e0,
        radius_x: float = 0.0e0,
        radius_y: float = 0.0e0,
        radius_z: float = 0.0e0,
        rp: float = 0.0e0,
        radius_xp: float = 0.0e0,
        radius_yp: float = 0.0e0,
        normalized_emittance: float = 0.0e0,
        normalized_emittance_x: float = 0.0e0,
        normalized_emittance_y: float = 0.0e0,
        x: float = 0.0e0,
        y: float = 0.0e0,
        z: float = 0.0e0,
        xp: float = 0.0e0,
        yp: float = 0.0e0,
    ):
        """
        Initialization of an electron beam.

        Parameters
        ----------
        current: float
            Beam current, [A]
        energy: float
            Beam energy, [MeV]
        radius: float
            Beam radius, [m]
        radius_x: float, optional
            Elliptical beam radius along the x-axis, [m]
        radius_y: float, optional
            Elliptical beam radius along the y-axis, [m]
        rp: float
            Beam radius prime, [rad]
        radius_xp: float, optional
            Elliptical beam radius prime along the x-axis, [rad]
        radius_yp: float, optional
            Elliptical beam radius prime along the y-axis, [rad]
        normalized_emittance: float
            Beam normalized emittance, [m*rad]
        normalized_emittance_x: float, optional
            Elliptical beam normalized emittance along the x-axis, [m*rad]
        normalized_emittance_y: float, optional
            Elliptical beam normalized emittance along the y-axis, [m*rad]

        x: float, optional
            Offset of the centroid along the x-axis, [m]
        xp: float, optional
            Centroid rotation in the z-x plane, [rad]
        y: float, optional
            Offset of the centroid along the y-axis, [m]
        yp: float, optional
            Centroid rotation in the z-y plane, [rad]
        """
        self.current = current
        self.energy = energy
        self.radius = radius
        self.rp = rp
        self.radius_x = radius_x
        self.radius_y = radius_y
        self.radius_xp = radius_xp
        self.radius_yp = radius_yp
        self.normalized_emittance = normalized_emittance
        self.normalized_emittance_x = normalized_emittance_x
        self.normalized_emittance_y = normalized_emittance_y
        if radius != 0.0e0:
            self.radius_x = radius
            self.radius_y = radius
        if rp != 0.0e0:
            self.radius_xp = rp
            self.radius_yp = rp
        if normalized_emittance != 0.0e0:
            self.normalized_emittance_x = normalized_emittance
            self.normalized_emittance_y = normalized_emittance

        self.x = x
        self.y = y
        self.xp = xp
        self.yp = yp

        self.gamma = gamma = self.energy / const.electron_mass_energy + 1
        self.beta = beta = np.sqrt(1 - 1 / (gamma * gamma))

        self.p = self.momentum = gamma * beta * const.electron_mass_energy
        self.px = self.p * self.radius_xp
        self.py = self.p * self.radius_yp
        self.pz = self.p
        self.description = ""

        self.type = type  # particles type
        self.n = 0.0  # quantity
        self.df = pd.DataFrame  # data frame
        self.radius_z = radius_z
        self.z = z
        self.total_charge = (
            (2 * type.charge / abs(type.charge) * self.current * self.radius_z / const.c / self.beta)
            if self.beta != 0.0e0
            else 0.0e0
        )

    @abstractmethod
    def _generate(self, n: int = 0, *, file_name: str = "", **kwargs) -> None:
        pass

    def generate(self, n: int = 0, *, file_name: str = "", **kwargs) -> None:
        if n < 0:
            raise ValueError("The number of particles (n) must be a positive number!")
        module_logger.info("Generate a %s beam with %d particles", self.distribution, n)
        self._generate(n, file_name=file_name, **kwargs)

    def __str__(self):
        return f"""Beam parameters:
            Type\t{self.type.name}
            Distribution\t{self.distribution}
            Particles\t{self.n}
            Current\t{self.current} A
            Energy\t{self.energy} MeV
            Total momentum\t{self.momentum} MeV/c
            Rel. factor\t{self.gamma}
            Radius x\t{self.radius_x / const.milli} mm
            Radius y\t{self.radius_y / const.milli} mm
            Radius z\t{self.radius_z} m
            Radius x prime\t{self.radius_xp / const.milli} mrad
            Radius y prime\t{self.radius_yp / const.milli} mrad
            Horizontal centroid position\t{self.x / const.milli} mm
            Vertical centroid position\t{self.y / const.milli} mm
            Horizontal centroid angle\t{self.xp / const.milli} mrad
            Vertical centroid angle\t{self.yp / const.milli} mrad
            Normalized emittance x\t{self.normalized_emittance_x / const.milli / const.milli} mm*mrad
            Normalized emittance y\t{self.normalized_emittance_y / const.milli / const.milli} mm*mrad
        """



from setuptools import find_packages, setup

from redpic.core import config

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements/base.txt", "r", encoding="utf-8") as fh:
    required = fh.read().splitlines()

setup(
    name=config.NAME,
    version=config.VERSION,
    author=config.AUTHOR,
    author_email=config.AUTHOR_EMAIL,
    description=config.DOC,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url=config.URL,
    packages=find_packages(),
    install_requires=required,
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Physics",
    ],
    project_urls={
        "Bug Reports": "https://github.com/fuodorov/redpic/issues",
        "PyPi": "https://pypi.org/project/redpic/",
        "Documentation": "https://fuodorov.github.io/redpic/",
        "Source": "https://github.com/fuodorov/redpic/",
    },
)



import logging

import numpy as np
import pandas as pd

from redpic import constants as const
from redpic.beam.base import BaseBeam, BeamDistributionType

module_logger = logging.getLogger(__name__)


class AstraBeam(BaseBeam):
    distribution = BeamDistributionType.ASTRA

    def _read_astra_particles(self, file_name: str) -> tuple:
        module_logger.info("Read particles from Astra file")
        cols = [
            "x",
            "y",
            "z",
            "px",
            "py",
            "pz",
            "clock",
            "charge",
            "id",
            "flag",
        ]  # m    m    m    eV/c  eV/c  eV/c  ns       nC
        df = pd.read_csv(file_name, header=None, sep=r"\s+", names=cols, dtype="float32")
        df = df[df.flag != -15]  # ignore the lost particles
        df["px"] = df["px"] / const.mega  # MeV/c
        df["py"] = df["py"] / const.mega  # MeV/c
        df0 = df.head(1)  # remove the reference particle
        df = df.drop(df0.index)
        z0 = df0.z.values[0]
        pz0 = df0.pz.values[0]
        df["z"] = z0 + df["clock"] * 1e-9 * const.c  # m
        df["pz"] = (pz0 + df["pz"]) / const.mega  # MeV/c
        return df["x"], df["y"], df["z"], df["px"], df["py"], df["pz"]

    def _generate(self, n: int = 0, *, file_name: str = "", **kwargs) -> None:
        module_logger.info("Particle loading from Astra %s file", file_name)
        x, y, z, px, py, pz = self._read_astra_particles(file_name)
        self.n = int(len(x))
        self.df = pd.DataFrame(
            np.transpose(np.vstack((x, y, z, px, py, pz))), columns=["x", "y", "z", "px", "py", "pz"]
        )
